import os
from collections.abc import Iterable
from .mapper_interface import MapperInterface
import csv
from typing import Union
import importlib.resources
from icdmappings import data_files


class ICD9toICD10(MapperInterface):
    """
    Maps icd9 codes into icd10.
    
    Source of mapping: https://www.nber.org/research/data/icd-9-cm-and-icd-10-cm-and-icd-10-pcs-crosswalk-or-general-equivalence-mappings
    """
    def __init__(self):
        self.filename = "icd9toicd10cmgem.csv"
        self._setup()

    def _setup(self):
        self.icd9_to_icd10 = self._parse_file(self.filename)

    def _map_single(self, icd9code : str):
            
        return self.icd9_to_icd10.get(icd9code)

    def map(self, icd9code : Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Given an icd9 code, returns the corresponding icd10 code.

            Parameters
            ----------

            code : str | pd.Series
                icd9 code

            Returns:
                icd10 code or None when the mapping is not possible
            """
            
            if isinstance(icd9code, str):
                return  self._map_single(icd9code)
            
            elif isinstance(icd9code, Iterable):
                return [self._map_single(c) for c in icd9code]

    def _parse_file(self, filename : str):

        mapping = {}

        with importlib.resources.open_text(data_files, filename) as csvfile:
            reader = csv.reader(csvfile, quotechar='"')
            headers = next(reader)

            for row in reader:
                icd9, icd10 = row[0], row[1]

                mapping[icd9] = icd10
        return mapping




from typing import List, Union
import os
from collections.abc import Iterable
import csv
import importlib.resources
from icdmappings import data_files

class ICD9toCCI:
        """
        Classifies ICD9 diagnostic codes into Chronic (True) or not Chronic (False).
        
        source of mapping: https://www.hcup-us.ahrq.gov/toolssoftware/chronic/chronic.jsp
        """
        def __init__(self):
            self.filename = "cci2015.csv"
            self.icd9_to_cci = None # will be filled by self._setup() {icd9code:cci,...icd9code:cci}
            self._setup()


        def _setup(self):
            # creates self.chapters_num, self.chapters_char, self.bins
            self.icd9_to_cci = self._parse_file(self.filename)

        def _map_single(self, icd9code : str) -> str:
             return self.icd9_to_cci.get(icd9code)


        def map(self, icd9code : Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Given an icd9 code, returns the corresponding Chronic classification
            (True for chronic, and False for not-chronic)

            Parameters
            ----------
            code : str or Iterable
                icd9 code or iterable of icd9 codes in string format.

            Returns
            -------
            True: When the code is chronic
            False: when the code is not chronic
            None: code is not recognizable
            """

            if isinstance(icd9code, str):
                return self._map_single(icd9code)
            elif isinstance(icd9code, Iterable):
                return [self._map_single(c) for c in icd9code]


        def _parse_file(self, filename : str):
            with importlib.resources.open_text(data_files, filename) as csvfile:
                reader = csv.reader(csvfile, quotechar="'")
                headers = next(reader)

                cci_to_bool = {'1':True,'0':False}

                mapping = {}

                for row in reader:
                    icd9_code = row[0].strip()
                    cci = cci_to_bool[row[2]]
                    mapping[icd9_code] = cci

            return mapping


from typing import List,Union
import os
from collections.abc import Iterable
from .mapper_interface import MapperInterface
import bisect
import csv
import importlib.resources
from icdmappings import data_files

class ICD9toChapters(MapperInterface):
        """
        Maps icd9 diagnostic codes to icd9 chapters.
        
        source of mapping: https://icd.codes/icd9cm
        """
        def __init__(self):
            self.filename = "icd9-CM-code-chapter-en=PT.csv"
            self._setup()

        def _setup(self):
            self.bins, self.char_mapping = self._parse_file(self.filename)
        
        def _map_single(self, icd9code : str) -> str:
            """
            Maps a single ICD9 code in string format to its corresponding chapter.

            Parameters
            ----------
            icd9code : str
                ICD9 code to be mapped.

            Returns
            -------
            chapter : str
                Corresponding icd9 chapter or None if mapping is unsuccessful.
            """

            if not icd9code: # empty string or None
                return None

            if not isinstance(icd9code, str): # has to be string
                return None
            
            if icd9code[0] in ['E','V']:
                return self.char_mapping[icd9code[0]]
            else: # numerical code
                try:
                    code3digits = int(icd9code[:3])
                    bin = self._get_bin(code3digits,self.bins)
                    if bin is not None:
                        return str(bin)
                    return None
                except:
                    return None


        def map(self, icd9code : Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Maps an ICD9 code or an Iterable of ICD9 codes to their corresponding chapters.

            Parameters
            ----------
            code : str | Iterable

            Returns
            -------
            chapter : str | Iterable
                Corresponding icd9 chapter(s) or None when mapping is unsuccessful.
            """
            if isinstance(icd9code,str):
                return self._map_single(icd9code)
            
            elif isinstance(icd9code, Iterable):
                return [self._map_single(code) for code in icd9code]

        def _get_bin(self, number : Union[int, Iterable], bins : List):
            if isinstance(number,int):
                return bisect.bisect(bins, number)
            elif isinstance(number,Iterable):
                return [bisect.bisect(bins, num) if isinstance(num,int) else None for num in number]
            return None
            

        def _parse_file(self, filename : str):
            """
            Preprocessing of bins to optimize assignment speed during mapping.
            1. Separate chapters into numeric codes or alpha codes (There are two chapters considered alpha because they start with "E" or "V")
            2. Create self.bins, which contains starting code ranges of each chapter
            """

            with importlib.resources.open_text(data_files, filename) as csvfile:
                reader = csv.reader(csvfile, quotechar="'")
                headers = next(reader)

                bins = list()

                for row in reader:
                    code_range = row[1]

                    if code_range[0] in ['E','V']:
                        continue

                    start,end = code_range.split('-')

                    bins.append(int(start))
                bins.append(int(end)+1)
            
            char_mapping = {'E':'19','V':'18'}
            return bins, char_mapping


from abc import ABC, abstractclassmethod
from collections.abc import Iterable
from typing import Union

class ICDValidatorInterface(ABC):
    """
    Abstract class for mapping classes.
    """

    def __init__(self):
        pass
    
    @abstractclassmethod
    def validate_diagnostics(self, 
            codes : Union[str, Iterable],
            ) -> Union[bool, Iterable]:
        """
        Returns True of False if the code is a valid diagnostic code.
        """
        pass

    @abstractclassmethod
    def validate_procedures(self, 
            codes : Union[str, Iterable],
            ) -> Union[bool, Iterable]:
        """
        Returns True of False if the code is a valid procedure code.
        """
        pass


from typing import List, Union
import os
from collections.abc import Iterable
import csv
import importlib.resources
from icdmappings import data_files

class ICD10toCCIR:
        """
        Classifies ICD10 diagnostic codes into chronic (True) or not chronic (False).
        
        source of mapping: https://www.hcup-us.ahrq.gov/toolssoftware/chronic/chronic.jsp
        """
        def __init__(self):
            self.filename = "CCIR_v2023-1.csv"
            self.icd10_to_ccir = self._parse_file(self.filename) # {icd10code:cci,...icd10code:cci}

        def _map_single(self, icd10code : str) -> str:
             return self.icd10_to_ccir.get(icd10code)


        def map(self, icd10code : Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Given an icd10 code, returns the corresponding Chronic classification
            (True for chronic, and False for not-chronic)

            Parameters
            ----------
            code : str or Iterable
                icd9 code or iterable of icd10 codes in string format.

            Returns
            -------
            True: When the code is chronic
            False: when the code is not chronic
            None: code is not recognizable
            """

            if isinstance(icd10code, str):
                return self._map_single(icd10code)
            elif isinstance(icd10code, Iterable):
                return [self._map_single(c) for c in icd10code]


        def _parse_file(self, filename : str):
            with importlib.resources.open_text(data_files, filename) as csvfile:
                reader = csv.reader(csvfile, quotechar="'")
                headers = next(reader)
                more_headers = next(reader)
                and_more_headers = next(reader)

                # 9 means not determined
                ccir_to_bool = {'1':True, '0':False, '9': None}

                mapping = {}

                for row in reader:
                    icd10_code = row[0].strip()
                    ccir = ccir_to_bool[row[-1]]
                    mapping[icd10_code] = ccir

            return mapping


import os
from collections.abc import Iterable
from .mapper_interface import MapperInterface
import csv
from typing import Union
import importlib.resources
from icdmappings import data_files

class ICD10toICD9(MapperInterface):
    """
    Maps icd10 codes to icd9.
    
    Source of mapping: https://www.nber.org/research/data/icd-9-cm-and-icd-10-cm-and-icd-10-pcs-crosswalk-or-general-equivalence-mappings
    """
    def __init__(self):
        self.filename = "icd10cmtoicd9gem.csv"
        self._setup()

    def _setup(self):
        self.icd10_to_icd9 = self._parse_file(self.filename)


    def _map_single(self, icd10code : str):
        return self.icd10_to_icd9.get(icd10code)

    def map(self, icd10code : Union[str, Iterable]) -> Union[str, Iterable]:
        """
        Given an icd10 code, returns the corresponding icd9 code.

        Parameters
        ----------

        code : str | Iterable
            icd10 code

        Returns:
            icd9 code or None when the mapping is not possible
        """
            
        if isinstance(icd10code, str):
            return self._map_single(icd10code)

        elif isinstance(icd10code, Iterable):
            return [self._map_single(c) for c in icd10code]
        
        return None


    def _parse_file(self, filename : str):

        mapping = {}

        with importlib.resources.open_text(data_files, filename) as csvfile:
            reader = csv.reader(csvfile, quotechar='"')
            headers = next(reader)

            for row in reader:
                icd10, icd9 = row[0].strip(), row[1].strip()

                mapping[icd10] = icd9
        return mapping


from typing import Union
from collections.abc import Iterable
from .mapper_interface import MapperInterface
import json
import importlib.resources
from icdmappings import data_files
from icdmappings.data_files import ICD10_CM_Chapters

class ICD10toChapters(MapperInterface):
        """
        Maps ICD-10 CM diagnostic codes to ICD-10 chapters.
        
        source of mapping: https://icd.who.int/browse10/2010/en
        """
        def __init__(self):
            self.filename = "chapter_lookup.json"
            self._setup()

        def _setup(self):
            self.chapter_lookup = self._parse_file(self.filename)
        
        def _map_single(self, icd10code : str) -> str:
            """
            Maps a single ICD-10 CM code in string format to its corresponding chapter.

            Parameters
            ----------
            icd10code : str
                ICD-10 CM code to be mapped.

            Returns
            -------
            chapter : str
                Corresponding icd9 chapter or None if mapping is unsuccessful.
            """

            if not isinstance(icd10code, str): # has to be string
                return None
            

            letter = icd10code[0]
            if letter not in self.chapter_lookup:
                return None
            
            if len(icd10code) < 3:
                return None
            try:
                num = int(icd10code[1:3])
            except:
                return None
            
            for candidate_chapter in self.chapter_lookup[letter]:
                # use python's bisect to check if number is within bounds
                if num >= candidate_chapter['range'][0] and num <= candidate_chapter['range'][1]:
                    return candidate_chapter['chapter']
            
            return None



        def map(self, icd10code : Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Maps an ICD-10 CM code or an Iterable of ICD-10 CM codes to their corresponding chapters.

            Parameters
            ----------
            code : str | Iterable

            Returns
            -------
            chapter : str | Iterable
                Corresponding icd10 chapter(s) or None when mapping is unsuccessful.
            """

            if isinstance(icd10code, str):
                return self._map_single(icd10code)
            
            elif isinstance(icd10code, Iterable):
                return[self._map_single(code) for code in icd10code]
            
            return None
          
            

        def _parse_file(self, filename : str):
            """
            Preprocessing of bins to optimize assignment speed during mapping.
            1. Separate chapters into numeric codes or alpha codes (There are two chapters considered alpha because they start with "E" or "V")
            2. Create self.bins, which contains starting code ranges of each chapter
            """

            with importlib.resources.open_text(ICD10_CM_Chapters, filename) as jsonfile:
                chapter_lookup = json.load(jsonfile)

            return chapter_lookup


from abc import ABC, abstractclassmethod
from typing import List, Union
from collections.abc import Iterable

class MapperInterface(ABC):
    """
    Abstract class for mapping classes.
    """

    def __init__(self):
        pass
    
    @abstractclassmethod
    def map(self, 
            codes : Union[str, Iterable],
            ) -> Union[str, Iterable]:
        """
        Maps input codes to target encoding. 
        If unsuccessful mapping, returns None.
        """
        pass


from .mapper_interface import MapperInterface
from .icd9_to_ccs import ICD9toCCS
from .icd9_to_chapters import ICD9toChapters
from .icd9_to_cci import ICD9toCCI
from .icd9_to_icd10 import ICD9toICD10
from .icd10_to_icd9 import ICD10toICD9
from .icd10_to_chapters import ICD10toChapters
from .icd10_to_ccsr import ICD10toCCSR
from .icd10_to_ccir import ICD10toCCIR


__all__ = ["ICD9toCCS", "ICD9toChapters", "ICD9toCCI", "ICD9toICD10",  "ICD10toICD9", "ICD10toChapters", "ICD10toCCSR", "ICD10toCCIR"]


from .mappers import * 
from .validators import ICD9Validator
from collections.abc import Iterable
from typing import Union
class Mapper():

    def __init__(self):

        # mappers
        self.icd9_to_cci = ICD9toCCI()
        self.icd9_to_ccs = ICD9toCCS()
        self.icd9_to_chapters = ICD9toChapters()
        self.icd9_to_icd10 = ICD9toICD10()
        self.icd10_to_icd9 = ICD10toICD9()
        self.icd10_to_chapters = ICD10toChapters()
        self.icd10_to_ccsr = ICD10toCCSR()
        self.icd10_to_ccir = ICD10toCCIR()

        self._internal_mapping = {
                'icd9':{'cci':self.icd9_to_cci,
                        'ccs':self.icd9_to_ccs,
                        'chapter':self.icd9_to_chapters,
                        'icd10':self.icd9_to_icd10
                        },
                'icd10':{'icd9':self.icd10_to_icd9,
                         'chapter':self.icd10_to_chapters,
                         'ccsr':self.icd10_to_ccsr,
                         'ccir': self.icd10_to_ccir
                         }
                        }

        # validators

        self.icd9_validator = ICD9Validator()

        self._internal_validators = {
            'icd9': self.icd9_validator
        }

    def show_mappers(self):
        print('Here are the available mappers\n')
        for _from in self._internal_mapping:
            print('From ' + _from + ' to:')
            for _to in self._internal_mapping[_from]:
                print('\t- ' + _to)
    
    def show_validators(self):
        return list(self._internal_validators.keys())
    
    def validate_diagnostics(self, codes : Union[str,Iterable], category : str):

        validator = self._get_validator(category)
        
        return validator.validate_diagnostics(codes)
    
    def validate_procedures(self, codes : Union[str,Iterable], category : str):

        validator = self._get_validator(category)
        
        return validator.validate_procedures(codes)

    def _get_validator(self, category : str):
        validator = self._internal_validators.get(category)

        if validator is None:
            raise ValueError(f"Category must be one of the following {str(self._internal_validators.keys())}")
        return validator

    def map(self, 
            codes : Union[str, Iterable],
            source : str,
            target : str):
    
        _source = self._internal_mapping.get(source)
    
        if _source is None:
            raise ValueError(f'There\'s no mapper that starts from {source} codes. Available starting codes are: {str(list(self._internal_mapping.keys()))}. Use .show_mappers() for more info.')
        
        mapper = _source.get(target)
    
        if mapper is None:
            raise ValueError(f'There\'s no mapper that maps from {source} to {target}. Available mappers can only map from {source} to {str(list(_source.keys()))}. Use .show_mappers() for more info.')
        
        mapping = mapper.map(codes)
        
        return mapping


from typing import List, Union
import os
from collections.abc import Iterable
import re
from .mapper_interface import MapperInterface
import importlib.resources
from icdmappings import data_files

class ICD9toCCS(MapperInterface):
        """
        Maps icd9 codes to CCS groups
        
        source of mapping: https://www.hcup-us.ahrq.gov/toolssoftware/ccs/ccs.jsp
        """
        def __init__(self):
            self.filename = "CCS-SingleDiagnosisGrouper.txt"
            self._setup()


        def _setup(self):
            # {ccs:[icd9,...],..., ccs:[icd9,...]}
            self.ccs_to_icd9 = self._parse_file(self.filename)

            # (inverse mapping): {icd9:ccs,..., icd9:ccs}
            self.icd9_to_ccs = self.ccs_to_icd9 = {self.ccs_to_icd9[ccs][i]:ccs for ccs in self.ccs_to_icd9 for i in range(len(self.ccs_to_icd9[ccs]))} 
        
        def _parse_file(self, filename : str):
            with importlib.resources.open_text(data_files, filename) as f:
                content = f.read()
            return self._get_codes(content)
        
        def _map_single(self, icd9code : str) -> str:
            return self.icd9_to_ccs.get(icd9code)

        def map(self,icd9code: Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Given an icd9 code, returns the corresponding CCS code.
            If input is Iterable returns a list of codes. 
            
            Parameters
            ----------
            
            code : str | Iterable
                icd9 code
            
            Returns:
                ccs code or None when the mapping is not possible
            """
            
            if isinstance(icd9code, str):
                return self._map_single(icd9code)
            
            elif isinstance(icd9code, Iterable):
                return[self._map_single(code) for code in icd9code]
            
            return None

        def _get_codes(self, content : str):
            """
            Parses the file and returns a dictionary with the following structure:
            {ccs_code:[icd9_codes],...,ccs_code:[icd9_codes]}

            """

            groups = re.findall(r'(\d+\s+[A-Z].*(\n.+)+)',content)
            """
            Unfortunately we need this function because this regex isn't perfect

            Rules:
            # ccs code is always first element
            # always ignore empty strings
            # while in the first line, gotta wait for a \n inside a string
            # after the first \n we have icd9 codes.
            # some icd9 codes will have \n as they are the last code before a newline and the next ccs code
            # some strings may be just \n without any text attached



            Returns
            -------

            data : dict
                {ccs_code:[icd9_codes],...,ccs_code:[icd9_codes}
            """


            data = {}

            for group in groups:

                group = group[0]
                tokens = group.split(' ')
                ccs_code = None

                is_first_tok =True # first token is a ccs code
                is_first_line = True #ignore all tokens in the first line (except the first which is a ccs code)

                for tok in tokens:
                    if is_first_tok: # first token is always the ccs code
                        ccs_code = tok
                        data[ccs_code] = []

                        is_first_tok = False
                        continue

                    if tok == '': #ignore empty strings resulted from .split
                        continue

                    if '\n' in tok:

                        if tok == '\n':
                            if is_first_line: #We are not in the first line anymore
                                is_first_line=False
                            continue
                        else:
                            if is_first_line: #We are not in the first line anymore
                                is_first_line=False
                                continue 
                            else:
                                tok = tok.replace('\n','') # code with a \n attached. clean it

                    elif is_first_line: # Ignore everything in the first line
                        continue

                    # this token wasn't ignored in the previous steps. save it as a icd9 code
                    data[ccs_code].append(tok)
            return data


from .icd_validator_interface import ICDValidatorInterface
from typing import List, Union
import os
from collections.abc import Iterable
import importlib.resources
from icdmappings import data_files
from icdmappings.data_files import ICD_9_CM_v32_master_descriptions


class ICD9Validator(ICDValidatorInterface):
        """
        Checks if a code is icd9-cm version 32 compliant.
        """
        
        def __init__(self):
            self.diagnostics_filename = 'CMS32_DESC_LONG_DX.txt'
            self.procedures_filename = 'CMS32_DESC_LONG_SG.txt'
            self._setup()
        
        def _setup(self):
            self.diagnostics = self._parse_file(self.diagnostics_filename)
            self.procedures = self._parse_file(self.procedures_filename)
            pass
        

        def validate_diagnostics(self, 
                                codes : Union[str,Iterable],
                                ) -> Union[bool, Iterable]:
            """validates if a code or iterable of codes are valid diagnostics.
            If iterable is numpy or pd.Series, returns the same type. ALl other iterables are returned as List.

            Args:
                codes (str | List | pd.Series | np.ndarray): _description_

            Raises:
                ValueError: _description_

            Returns:
                bool | List | pd.Series | np.ndarray: _description_
            """
            if codes is None:
                return None
            if isinstance(codes,str):
                return codes in self.diagnostics.keys()
            elif isinstance(codes,Iterable):
                valid = [code in self.diagnostics.keys() if code is not None else None for code in codes]
                return valid
            raise TypeError('Expects a string or iterable of strings as codes.')
        
        def validate_procedures(self,
                               codes : Union[str, Iterable],
                              ) -> Union[bool, Iterable]:
            
            if codes is None:
                return None
            if isinstance(codes,str):
                return codes in self.procedures.keys()
            elif isinstance(codes,Iterable):
                valid = [code in self.procedures.keys() if code is not None else None for code in codes]
                return valid
            raise TypeError('Expects a string or iterable of strings as codes.')

        def _parse_file(self,filename : str):
            """Parses a data file of icd9 codes. and returns a dictionary of codes and descriptions.

            Args:
                f (str): file path to read from.

            Returns:
                data[code]: description
            """
            data = dict()
            with importlib.resources.open_text(ICD_9_CM_v32_master_descriptions, filename,encoding='latin-1') as f:
                for line in f:
                    code,desc = line.split(sep=' ', maxsplit=1)
                    data[code] = desc.strip()
            return data


from typing import Union
from collections.abc import Iterable
from .mapper_interface import MapperInterface
import json
import importlib.resources
from icdmappings.data_files import ICD10_CM_CCSR

class ICD10toCCSR(MapperInterface):
        """
        Maps ICD-10 CM diagnostic codes to CCS(R) chapters.
        
        source of mapping: https://hcup-us.ahrq.gov/toolssoftware/ccsr/ccs_refined.jsp
        """
        def __init__(self):
            self.filename = "dx_cat1_mapping.json"
            self._setup()

        def _setup(self):
            self.ccsr_lookup = self._parse_file(self.filename)
        
        def _map_single(self, icd10code : str) -> str:
            """
            Maps a single ICD-10 CM code in string format to its 
            corresponding CCS(R) code (Category 1).

            Why map to Category 1 of CCS(R)? Some ICD10-CM codes can map to 
            several CCS(R) categories, when this happens it is because 
            the condition can be specified with greater detail (Categories 2-6).

            By default, for now, we will map to Category 1, which is probably 
            enough to satisfy most use cases.

            Parameters
            ----------
            icd10code : str
                ICD-10 CM code to be mapped.

            Returns
            -------
            ccsr : str
                Corresponding CCS(R) code (Category 1) or None if mapping is unsuccessful.
            """

            if not isinstance(icd10code, str): # has to be string
                return None
            
            return self.ccsr_lookup.get(icd10code)



        def map(self, icd10code : Union[str, Iterable]) -> Union[str, Iterable]:
            """
            Maps an ICD-10 CM code or an Iterable of ICD-10 CM codes to their corresponding chapters.

            Parameters
            ----------
            code : str | Iterable

            Returns
            -------
            chapter : str | Iterable
                Corresponding icd10 chapter(s) or None when mapping is unsuccessful.
            """

            if isinstance(icd10code, str):
                return self._map_single(icd10code)
            
            elif isinstance(icd10code, Iterable):
                return[self._map_single(code) for code in icd10code]
            
            return None
          
            

        def _parse_file(self, filename : str):
            """
            Preprocessing of bins to optimize assignment speed during mapping.
            1. Separate chapters into numeric codes or alpha codes (There are two chapters considered alpha because they start with "E" or "V")
            2. Create self.bins, which contains starting code ranges of each chapter
            """

            with importlib.resources.open_text(ICD10_CM_CCSR, filename) as jsonfile:
                ccsr_lookup = json.load(jsonfile)

            return ccsr_lookup